@@ -29,18 +29,23 @@
 #include "config.h"
 #endif
 
+#ifdef ENABLE_HTTP_PROXY
+
 #include "syshead.h"
 
 #include "common.h"
-#include "buffer.h"
 #include "misc.h"
-#include "io.h"
+#include "win32.h"
 #include "socket.h"
 #include "fdmisc.h"
 #include "proxy.h"
+#include "ntlm.h"
 
 #include "memdbg.h"
 
+/* cached proxy username/password */
+static struct user_pass static_proxy_user_pass;
+
 static bool
 recv_line (socket_descriptor_t sd,
 	   char *buf,
@@ -77,7 +82,7 @@ recv_line (socket_descriptor_t sd,
 
       status = select (sd + 1, &reads, NULL, NULL, &tv);
 
-      GET_SIGNAL (*signal_received);
+      get_signal (signal_received);
       if (*signal_received)
 	goto error;
 
@@ -185,8 +190,8 @@ send_crlf (socket_descriptor_t sd)
   return send_line_crlf (sd, "");
 }
 
-static uint8_t *
-make_base64_string (const uint8_t *str)
+uint8_t *
+make_base64_string2 (const uint8_t *str, int src_len, struct gc_arena *gc)
 {
   static const char base64_table[] =
     "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
@@ -194,12 +199,11 @@ make_base64_string (const uint8_t *str)
   uint8_t *buf;
   const uint8_t *src;
   uint8_t *dst;
-  int bits, data, src_len, dst_len;
+  int bits, data, dst_len;
 
   /* make base64 string */
-  src_len = strlen (str);
   dst_len = (src_len + 2) / 3 * 4;
-  buf = gc_malloc (dst_len + 1);
+  buf = gc_malloc (dst_len + 1, false, gc);
   bits = data = 0;
   src = str;
   dst = buf;
@@ -209,8 +213,7 @@ make_base64_string (const uint8_t *str)
 	{
 	  data = (data << 8) | *src;
 	  bits += 8;
-	  if (*src != 0)
-	    src++;
+	  src++;
 	}
       *dst++ = base64_table[0x3F & (data >> (bits - 6))];
       bits -= 6;
@@ -228,72 +231,69 @@ make_base64_string (const uint8_t *str)
   return buf;
 }
 
-static const char *
-username_password_as_base64 (const struct http_proxy_info *p)
+uint8_t *
+make_base64_string (const uint8_t *str, struct gc_arena *gc)
 {
-  struct buffer out = alloc_buf_gc (strlen (p->username) + strlen (p->password) + 2);
-  ASSERT (strlen (p->username) > 0);
-  buf_printf (&out, "%s:%s", p->username, p->password);
-  return make_base64_string (BSTR (&out));
+  return make_base64_string2 (str, strlen (str), gc);
 }
 
-void
-init_http_proxy (struct http_proxy_info *p,
-		 const char *server,
-		 int port,
-		 bool retry,
-		 const char *auth_method,
-		 const char *auth_file)
+static const char *
+username_password_as_base64 (const struct http_proxy_info *p,
+			     struct gc_arena *gc)
 {
-  CLEAR (*p);
-  ASSERT (server);
-  ASSERT (legal_ipv4_port (port));
+  struct buffer out = alloc_buf_gc (strlen (p->up.username) + strlen (p->up.password) + 2, gc);
+  ASSERT (strlen (p->up.username) > 0);
+  buf_printf (&out, "%s:%s", p->up.username, p->up.password);
+  return make_base64_string (BSTR (&out), gc);
+}
 
-  strncpynt (p->server, server, sizeof (p->server));
-  p->port = port;
-  p->retry = retry;
-  p->auth_method = HTTP_AUTH_NONE;
+struct http_proxy_info *
+new_http_proxy (const struct http_proxy_options *o,
+		struct gc_arena *gc)
+{
+  struct http_proxy_info *p;
+  ALLOC_OBJ_CLEAR_GC (p, struct http_proxy_info, gc);
+
+  if (!o->server)
+    msg (M_FATAL, "HTTP_PROXY: server not specified");
+
+  ASSERT (legal_ipv4_port (o->port));
+
+  p->options = *o;
 
   /* parse authentication method */
-  if (auth_method)
+  p->auth_method = HTTP_AUTH_NONE;
+  if (o->auth_method_string)
     {
-      if (!strcmp (auth_method, "none"))
+      if (!strcmp (o->auth_method_string, "none"))
 	p->auth_method = HTTP_AUTH_NONE;
-      else if (!strcmp (auth_method, "basic"))
+      else if (!strcmp (o->auth_method_string, "basic"))
 	p->auth_method = HTTP_AUTH_BASIC;
+      else if (!strcmp (o->auth_method_string, "ntlm"))
+	p->auth_method = HTTP_AUTH_NTLM;
       else
-	msg (M_FATAL, "ERROR: unknown HTTP authentication method: '%s' -- only the 'none' or 'basic' methods are currently supported",
-	     auth_method);
+	msg (M_FATAL, "ERROR: unknown HTTP authentication method: '%s' -- only the 'none', 'basic', or 'ntlm' methods are currently supported",
+	     o->auth_method_string);
     }
 
-  /* only basic authentication supported so far */
-  if (p->auth_method == HTTP_AUTH_BASIC)
+  /* only basic and NTLM authentication supported so far */
+  if (p->auth_method == HTTP_AUTH_BASIC || p->auth_method == HTTP_AUTH_NTLM)
     {
-      FILE *fp;
-      
-      if (!auth_file)
-	msg (M_FATAL, "ERROR: http proxy authentication requires a username/password file");
-
-      p->auth_method = HTTP_AUTH_BASIC;
-      warn_if_group_others_accessible (auth_file);
-      fp = fopen (auth_file, "r");
-      if (!fp)
-	msg (M_ERR, "Error opening http proxy auth_file: %s", auth_file);
-      
-      if (fgets (p->username, sizeof (p->username), fp) == NULL
-	  || fgets (p->password, sizeof (p->password), fp) == NULL)
-	msg (M_FATAL, "Error reading username and password (must be on two consecutive lines) from http proxy authfile: %s", auth_file);
-      
-      fclose (fp);
-      
-      chomp (p->username);
-      chomp (p->password);
-      
-      if (strlen (p->username) == 0)
-	msg (M_FATAL, "ERROR: username from http proxy authfile '%s' is empty", auth_file);
+      get_user_pass (&static_proxy_user_pass,
+		     o->auth_file,
+		     false,
+		     "HTTP Proxy",
+		     0);
+      p->up = static_proxy_user_pass;
     }
 
+#if !NTLM
+  if (p->auth_method == HTTP_AUTH_NTLM)
+    msg (M_FATAL, "Sorry, this version of " PACKAGE_NAME " was built without NTLM Proxy support.");
+#endif
+
   p->defined = true;
+  return p;
 }
 
 void
@@ -304,18 +304,34 @@ establish_http_proxy_passthru (struct http_proxy_info *p,
 			       struct buffer *lookahead,
 			       volatile int *signal_received)
 {
-  char buf[128];
+  struct gc_arena gc = gc_new ();
+  char buf[256];
+  char buf2[128];
+  char get[80];
   int status;
   int nparms;
 
   /* format HTTP CONNECT message */
-  openvpn_snprintf (buf, sizeof(buf), "CONNECT %s:%d HTTP/1.0", host, port);
+  openvpn_snprintf (buf, sizeof(buf), "CONNECT %s:%d HTTP/%s",
+		    host,
+		    port,
+		    p->options.http_version);
+
   msg (D_PROXY, "Send to HTTP proxy: '%s'", buf);
 
   /* send HTTP CONNECT message to proxy */
   if (!send_line_crlf (sd, buf))
     goto error;
 
+  /* send User-Agent string if provided */
+  if (p->options.user_agent)
+    {
+      openvpn_snprintf (buf, sizeof(buf), "User-Agent: %s",
+			p->options.user_agent);
+      if (!send_line_crlf (sd, buf))
+	goto error;
+    }
+
   /* auth specified? */
   switch (p->auth_method)
     {
@@ -324,25 +340,37 @@ establish_http_proxy_passthru (struct http_proxy_info *p,
 
     case HTTP_AUTH_BASIC:
       openvpn_snprintf (buf, sizeof(buf), "Proxy-Authorization: Basic %s",
-			username_password_as_base64 (p));
+			username_password_as_base64 (p, &gc));
       msg (D_PROXY, "Attempting Basic Proxy-Authorization");
-      msg (D_SHOW_KEYS, "Send to HTTP proxy: '%s'", buf);
-      sleep (1);
+      dmsg (D_SHOW_KEYS, "Send to HTTP proxy: '%s'", buf);
+      openvpn_sleep (1);
       if (!send_line_crlf (sd, buf))
 	goto error;
       break;
 
+#if NTLM
+    case HTTP_AUTH_NTLM:
+      openvpn_snprintf (buf, sizeof(buf), "Proxy-Authorization: NTLM %s",
+			ntlm_phase_1 (p, &gc));
+      msg (D_PROXY, "Attempting NTLM Proxy-Authorization phase 1");
+      dmsg (D_SHOW_KEYS, "Send to HTTP proxy: '%s'", buf);
+      openvpn_sleep (1);
+      if (!send_line_crlf (sd, buf))
+	goto error;
+      break;
+#endif
+
     default:
       ASSERT (0);
     }
 
   /* send empty CR, LF */
-  sleep (1);
+  openvpn_sleep (1);
   if (!send_crlf (sd))
     goto error;
 
   /* receive reply from proxy */
-  if (!recv_line (sd, buf, sizeof(buf), 5, true, NULL, signal_received))
+  if (!recv_line (sd, buf, sizeof(buf), p->options.timeout, true, NULL, signal_received))
     goto error;
 
   /* remove trailing CR, LF */
@@ -353,15 +381,105 @@ establish_http_proxy_passthru (struct http_proxy_info *p,
   /* parse return string */
   nparms = sscanf (buf, "%*s %d", &status);
 
+  /* check for a "407 Proxy Authentication Required" response */
+  if (nparms >= 1 && status == 407)
+    {
+      msg (D_PROXY, "Proxy requires authentication");
+
+      // check for NTLM
+      if (p->auth_method == HTTP_AUTH_NTLM)
+        {
+#if NTLM
+          // look for the phase 2 response
+
+          while (true)
+            {
+              if (!recv_line (sd, buf, sizeof(buf), p->options.timeout, true, NULL, signal_received))
+                goto error;
+              chomp (buf);
+              msg (D_PROXY, "HTTP proxy returned: '%s'", buf);
+
+              openvpn_snprintf (get, sizeof get, "%%*s NTLM %%%ds", sizeof (buf2) - 1);
+              nparms = sscanf (buf, get, buf2);
+              buf2[127] = 0; // we only need the beginning - ensure it's null terminated.
+
+              // check for "Proxy-Authenticate: NTLM TlRM..."
+              if (nparms == 1)
+                {
+                  // parse buf2
+                  msg (D_PROXY, "auth string: '%s'", buf2);
+                  break;
+                }
+            }
+          // if we are here then auth string was got
+          msg (D_PROXY, "Received NTLM Proxy-Authorization phase 2 response");
+
+          /* receive and discard everything else */
+          while (recv_line (sd, NULL, 0, p->options.timeout, true, NULL, signal_received))
+            ;
+
+          /* now send the phase 3 reply */
+
+          /* format HTTP CONNECT message */
+          openvpn_snprintf (buf, sizeof(buf), "CONNECT %s:%d HTTP/%s",
+			    host,
+			    port,
+			    p->options.http_version);
+
+          msg (D_PROXY, "Send to HTTP proxy: '%s'", buf);
+
+          /* send HTTP CONNECT message to proxy */
+          if (!send_line_crlf (sd, buf))
+            goto error;
+
+          /* send HOST etc, */
+          openvpn_sleep (1);
+          openvpn_snprintf (buf, sizeof(buf), "Host: %s", host);
+          msg (D_PROXY, "Send to HTTP proxy: '%s'", buf);
+          if (!send_line_crlf (sd, buf))
+            goto error;
+
+          openvpn_snprintf (buf, sizeof(buf), "Proxy-Authorization: NTLM %s",
+			    ntlm_phase_3 (p, buf2, &gc));
+          msg (D_PROXY, "Attempting NTLM Proxy-Authorization phase 3");
+          msg (D_PROXY, "Send to HTTP proxy: '%s'", buf);
+          openvpn_sleep (1);
+          if (!send_line_crlf (sd, buf))
+	    goto error;
+          // ok so far...
+          /* send empty CR, LF */
+          openvpn_sleep (1);
+          if (!send_crlf (sd))
+            goto error;
+
+          /* receive reply from proxy */
+          if (!recv_line (sd, buf, sizeof(buf), p->options.timeout, true, NULL, signal_received))
+            goto error;
+
+          /* remove trailing CR, LF */
+          chomp (buf);
+
+          msg (D_PROXY, "HTTP proxy returned: '%s'", buf);
+
+          /* parse return string */
+          nparms = sscanf (buf, "%*s %d", &status);
+#else
+	  ASSERT (0); // No NTLM support
+#endif
+	}
+      else goto error;
+    }
+
+
   /* check return code, success = 200 */
-  if (nparms != 1 || status != 200)
+  if (nparms < 1 || status != 200)
     {
       msg (D_LINK_ERRORS, "HTTP proxy returned bad status");
-#if 0
+#if 0 
       /* DEBUGGING -- show a multi-line HTTP error response */
       while (true)
 	{
-	  if (!recv_line (sd, buf, sizeof (buf), 5, true, NULL, signal_received))
+	  if (!recv_line (sd, buf, sizeof (buf), p->options.timeout, true, NULL, signal_received))
 	    goto error;
 	  chomp (buf);
 	  msg (D_PROXY, "HTTP proxy returned: '%s'", buf);
@@ -371,7 +489,7 @@ establish_http_proxy_passthru (struct http_proxy_info *p,
     }
 
   /* receive line from proxy and discard */
-  if (!recv_line (sd, NULL, 0, 5, true, NULL, signal_received))
+  if (!recv_line (sd, NULL, 0, p->options.timeout, true, NULL, signal_received))
     goto error;
 
   /*
@@ -386,11 +504,17 @@ establish_http_proxy_passthru (struct http_proxy_info *p,
     msg (M_INFO, "HTTP PROXY: lookahead: %s", format_hex (BPTR (lookahead), BLEN (lookahead), 0));
 #endif
 
+  gc_free (&gc);
   return;
 
  error:
   /* on error, should we exit or restart? */
   if (!*signal_received)
-    *signal_received = (p->retry ? SIGUSR1 : SIGTERM);
+    *signal_received = (p->options.retry ? SIGUSR1 : SIGTERM); /* SOFT-SIGUSR1 -- HTTP proxy error */
+  gc_free (&gc);
   return;
 }
+
+#else
+static void dummy(void) {}
+#endif /* ENABLE_HTTP_PROXY */
