@@ -65,12 +65,12 @@ reliable_ack_read_packet_id (struct buffer *buf, packet_id_type *pid)
   if (buf_read (buf, &net_pid, sizeof (net_pid)))
     {
       *pid = ntohpid (net_pid);
-      msg (D_REL_DEBUG, "ACK read ID " packet_id_format " (buf->len=%d)",
+      dmsg (D_REL_DEBUG, "ACK read ID " packet_id_format " (buf->len=%d)",
 	   (packet_id_print_type)*pid, buf->len);
       return true;
     }
 
-  msg (D_REL_LOW, "ACK read ID FAILED (buf->len=%d)", buf->len);
+  dmsg (D_REL_LOW, "ACK read ID FAILED (buf->len=%d)", buf->len);
   return false;
 }
 
@@ -81,12 +81,12 @@ reliable_ack_acknowledge_packet_id (struct reliable_ack *ack, packet_id_type pid
   if (!reliable_ack_packet_id_present (ack, pid) && ack->len < RELIABLE_ACK_SIZE)
     {
       ack->packet_id[ack->len++] = pid;
-      msg (D_REL_DEBUG, "ACK acknowledge ID " packet_id_format " (ack->len=%d)",
+      dmsg (D_REL_DEBUG, "ACK acknowledge ID " packet_id_format " (ack->len=%d)",
 	   (packet_id_print_type)pid, ack->len);
       return true;
     }
 
-  msg (D_REL_LOW, "ACK acknowledge ID " packet_id_format " FAILED (ack->len=%d)",
+  dmsg (D_REL_LOW, "ACK acknowledge ID " packet_id_format " FAILED (ack->len=%d)",
        (packet_id_print_type)pid, ack->len);
   return false;
 }
@@ -96,6 +96,7 @@ bool
 reliable_ack_read (struct reliable_ack * ack,
 		   struct buffer * buf, const struct session_id * sid)
 {
+  struct gc_arena gc = gc_new ();
   int i;
   uint8_t count;
   packet_id_type net_pid;
@@ -120,15 +121,17 @@ reliable_ack_read (struct reliable_ack * ack,
       if (!session_id_defined (&session_id_remote) ||
 	  !session_id_equal (&session_id_remote, sid))
 	{
-	  msg (D_REL_LOW,
+	  dmsg (D_REL_LOW,
 	       "ACK read BAD SESSION-ID FROM REMOTE, local=%s, remote=%s",
-	       session_id_print (sid), session_id_print (&session_id_remote));
+	       session_id_print (sid, &gc), session_id_print (&session_id_remote, &gc));
 	  goto error;
 	}
     }
+  gc_free (&gc);
   return true;
 
 error:
+  gc_free (&gc);
   return false;
 }
 
@@ -157,7 +160,7 @@ reliable_ack_write (struct reliable_ack * ack,
       packet_id_type pid = ack->packet_id[i];
       packet_id_type net_pid = htonpid (pid);
       ASSERT (buf_write (&sub, &net_pid, sizeof (net_pid)));
-      msg (D_REL_DEBUG, "ACK write ID " packet_id_format " (ack->len=%d, n=%d)", (packet_id_print_type)pid, ack->len, n);
+      dmsg (D_REL_DEBUG, "ACK write ID " packet_id_format " (ack->len=%d, n=%d)", (packet_id_print_type)pid, ack->len, n);
     }
   if (n)
     {
@@ -183,13 +186,13 @@ reliable_ack_adjust_frame_parameters (struct frame* frame, int max)
 
 /* print a reliable ACK record coming off the wire */
 const char *
-reliable_ack_print (struct buffer* buf, bool verbose)
+reliable_ack_print (struct buffer *buf, bool verbose, struct gc_arena *gc)
 {
   int i;
   uint8_t n_ack;
   struct session_id sid_ack;
   packet_id_type pid;
-  struct buffer out = alloc_buf_gc (256);
+  struct buffer out = alloc_buf_gc (256, gc);
 
   buf_printf (&out, "[");
   if (!buf_read (buf, &n_ack, sizeof (n_ack)))
@@ -206,7 +209,7 @@ reliable_ack_print (struct buffer* buf, bool verbose)
       if (!session_id_read (&sid_ack, buf))
 	goto done;
       if (verbose)
-	buf_printf (&out, " sid=%s", session_id_print (&sid_ack));
+	buf_printf (&out, " sid=%s", session_id_print (&sid_ack, gc));
     }
 
  done:
@@ -273,7 +276,7 @@ reliable_send_purge (struct reliable *rel, struct reliable_ack *ack)
 	  struct reliable_entry *e = &rel->array[j];
 	  if (e->active && e->packet_id == pid)
 	    {
-	      msg (D_REL_DEBUG,
+	      dmsg (D_REL_DEBUG,
 		   "ACK received for pid " packet_id_format ", deleting from send buffer",
 		   (packet_id_print_type)pid);
 #if 0
@@ -281,7 +284,7 @@ reliable_send_purge (struct reliable *rel, struct reliable_ack *ack)
 	      {
 		if (e->next_try)
 		  {
-		    const interval_t wake = e->next_try - time(NULL);
+		    const interval_t wake = e->next_try - now;
 		    msg (M_INFO, "ACK " packet_id_format ", wake=%d", pid, wake);
 		  }
 	      }
@@ -295,9 +298,9 @@ reliable_send_purge (struct reliable *rel, struct reliable_ack *ack)
 
 /* print the current sequence of active packet IDs */
 static const char *
-reliable_print_ids (const struct reliable *rel)
+reliable_print_ids (const struct reliable *rel, struct gc_arena *gc)
 {
-  struct buffer out = alloc_buf_gc (512);
+  struct buffer out = alloc_buf_gc (256, gc);
   int i;
 
   buf_printf (&out, "[" packet_id_format "]", (packet_id_print_type)rel->packet_id);
@@ -314,6 +317,7 @@ reliable_print_ids (const struct reliable *rel)
 bool
 reliable_can_get (const struct reliable *rel)
 {
+  struct gc_arena gc = gc_new ();
   int i;
   for (i = 0; i < rel->size; ++i)
     {
@@ -321,7 +325,8 @@ reliable_can_get (const struct reliable *rel)
       if (!e->active)
 	return true;
     }
-  msg (D_REL_LOW, "ACK no free receive buffer available: %s", reliable_print_ids (rel));
+  dmsg (D_REL_LOW, "ACK no free receive buffer available: %s", reliable_print_ids (rel, &gc));
+  gc_free (&gc);
   return false;
 }
 
@@ -329,6 +334,7 @@ reliable_can_get (const struct reliable *rel)
 bool
 reliable_not_replay (const struct reliable *rel, packet_id_type id)
 {
+  struct gc_arena gc = gc_new ();
   int i;
   if (id < rel->packet_id)
     goto bad;
@@ -338,10 +344,12 @@ reliable_not_replay (const struct reliable *rel, packet_id_type id)
       if (e->active && e->packet_id == id)
 	goto bad;
     }
+  gc_free (&gc);
   return true;
 
  bad:
-  msg (D_REL_DEBUG, "ACK " packet_id_format " is a replay: %s", (packet_id_print_type)id, reliable_print_ids (rel));
+  dmsg (D_REL_DEBUG, "ACK " packet_id_format " is a replay: %s", (packet_id_print_type)id, reliable_print_ids (rel, &gc));
+  gc_free (&gc);
   return false;
 }
 
@@ -349,16 +357,21 @@ reliable_not_replay (const struct reliable *rel, packet_id_type id)
 bool
 reliable_wont_break_sequentiality (const struct reliable *rel, packet_id_type id)
 {
+  struct gc_arena gc = gc_new ();
+  int ret;
+
   if ((int)id < (int)rel->packet_id + rel->size)
     {
-      return true;
+      ret = true;
     }
   else
     {
-      msg (D_REL_LOW, "ACK " packet_id_format " breaks sequentiality: %s",
-	   (packet_id_print_type)id, reliable_print_ids (rel));
-      return false;
+      dmsg (D_REL_LOW, "ACK " packet_id_format " breaks sequentiality: %s",
+	   (packet_id_print_type)id, reliable_print_ids (rel, &gc));
+      ret = false;
     }
+  gc_free (&gc);
+  return ret;
 }
 
 /* grab a free buffer */
@@ -382,9 +395,11 @@ reliable_get_buf (struct reliable *rel)
 struct buffer *
 reliable_get_buf_output_sequenced (struct reliable *rel)
 {
+  struct gc_arena gc = gc_new ();
   int i;
   packet_id_type min_id = 0;
   bool min_id_defined = false;
+  struct buffer *ret = NULL;
 
   /* find minimum active packet_id */
   for (i = 0; i < rel->size; ++i)
@@ -402,13 +417,14 @@ reliable_get_buf_output_sequenced (struct reliable *rel)
 
   if (!min_id_defined || (int)(rel->packet_id - min_id) < rel->size)
     {
-      return reliable_get_buf (rel);
+      ret = reliable_get_buf (rel);
     }
   else
     {
-      msg (D_REL_LOW, "ACK output sequence broken: %s", reliable_print_ids (rel));
-      return NULL;
+      dmsg (D_REL_LOW, "ACK output sequence broken: %s", reliable_print_ids (rel, &gc));
     }
+  gc_free (&gc);
+  return ret;
 }
 
 /* get active buffer for next sequentially increasing key ID */
@@ -429,8 +445,9 @@ reliable_get_buf_sequenced (struct reliable *rel)
 
 /* return true if reliable_send would return a non-NULL result */
 bool
-reliable_can_send (const struct reliable *rel, time_t current)
+reliable_can_send (const struct reliable *rel)
 {
+  struct gc_arena gc = gc_new ();
   int i;
   int n_active = 0, n_current = 0;
   for (i = 0; i < rel->size; ++i)
@@ -439,14 +456,16 @@ reliable_can_send (const struct reliable *rel, time_t current)
       if (e->active)
 	{
 	  ++n_active;
-	  if (current >= e->next_try)
+	  if (now >= e->next_try)
 	    ++n_current;
 	}
     }
-  msg (D_REL_DEBUG, "ACK reliable_can_send active=%d current=%d : %s",
+  dmsg (D_REL_DEBUG, "ACK reliable_can_send active=%d current=%d : %s",
        n_active,
        n_current,
-       reliable_print_ids (rel));
+       reliable_print_ids (rel, &gc));
+
+  gc_free (&gc);
   return n_current > 0;
 }
 
@@ -472,14 +491,16 @@ reliable_unique_retry (struct reliable *rel, time_t retry)
 
 /* return next buffer to send to remote */
 struct buffer *
-reliable_send (struct reliable *rel, int *opcode, time_t current)
+reliable_send (struct reliable *rel, int *opcode)
 {
   int i;
   struct reliable_entry *best = NULL;
+  const time_t local_now = now;
+
   for (i = 0; i < rel->size; ++i)
     {
       struct reliable_entry *e = &rel->array[i];
-      if (e->active && current >= e->next_try)
+      if (e->active && local_now >= e->next_try)
 	{
 	  if (!best || e->packet_id < best->packet_id)
 	    best = e;
@@ -487,18 +508,18 @@ reliable_send (struct reliable *rel, int *opcode, time_t current)
     }
   if (best)
     {
-#if 1
+#ifdef EXPONENTIAL_BACKOFF
       /* exponential backoff */
-      best->next_try = reliable_unique_retry (rel, current + best->timeout);
+      best->next_try = reliable_unique_retry (rel, local_now + best->timeout);
       best->timeout *= 2;
 #else
       /* constant timeout, no backoff */
-      best->next_try = current + best->timeout;
+      best->next_try = local_now + best->timeout;
 #endif
       *opcode = best->opcode;
-      msg (D_REL_DEBUG, "ACK reliable_send ID " packet_id_format " (size=%d to=%d)",
+      dmsg (D_REL_DEBUG, "ACK reliable_send ID " packet_id_format " (size=%d to=%d)",
 	   (packet_id_print_type)best->packet_id, best->buf.len,
-	   (int)(best->next_try - current));
+	   (int)(best->next_try - local_now));
       return &best->buf;
     }
   return NULL;
@@ -506,16 +527,16 @@ reliable_send (struct reliable *rel, int *opcode, time_t current)
 
 /* schedule all pending packets for immediate retransmit */
 void
-reliable_schedule_now (struct reliable *rel, time_t current)
+reliable_schedule_now (struct reliable *rel)
 {
   int i;
-  msg (D_REL_DEBUG, "ACK reliable_schedule_now");
+  dmsg (D_REL_DEBUG, "ACK reliable_schedule_now");
   for (i = 0; i < rel->size; ++i)
     {
       struct reliable_entry *e = &rel->array[i];
       if (e->active)
 	{
-	  e->next_try = current;
+	  e->next_try = now;
 	  e->timeout = rel->initial_timeout;
 	}
     }
@@ -524,31 +545,35 @@ reliable_schedule_now (struct reliable *rel, time_t current)
 /* in how many seconds should we wake up to check for timeout */
 /* if we return BIG_TIMEOUT, nothing to wait for */
 interval_t
-reliable_send_timeout (const struct reliable *rel, time_t current)
+reliable_send_timeout (const struct reliable *rel)
 {
+  struct gc_arena gc = gc_new ();
   interval_t ret = BIG_TIMEOUT;
   int i;
+  const time_t local_now = now;
 
   for (i = 0; i < rel->size; ++i)
     {
       const struct reliable_entry *e = &rel->array[i];
       if (e->active)
 	{
-	  if (e->next_try <= current)
+	  if (e->next_try <= local_now)
 	    {
 	      ret = 0;
 	      break;
 	    }
 	  else
 	    {
-	      ret = min_int (ret, e->next_try - current);
+	      ret = min_int (ret, e->next_try - local_now);
 	    }
 	}
     }
-  msg (D_REL_DEBUG, "ACK reliable_send_timeout %d %s",
-       (int) ret,
-       reliable_print_ids (rel));
 
+  dmsg (D_REL_DEBUG, "ACK reliable_send_timeout %d %s",
+       (int) ret,
+       reliable_print_ids (rel, &gc));
+
+  gc_free (&gc);
   return ret;
 }
 
@@ -577,7 +602,7 @@ reliable_mark_active_incoming (struct reliable *rel, struct buffer *buf,
 	  e->opcode = opcode;
 	  e->next_try = 0;
 	  e->timeout = 0;
-	  msg (D_REL_DEBUG, "ACK mark active incoming ID " packet_id_format, (packet_id_print_type)e->packet_id);
+	  dmsg (D_REL_DEBUG, "ACK mark active incoming ID " packet_id_format, (packet_id_print_type)e->packet_id);
 	  return;
 	}
     }
@@ -607,7 +632,7 @@ reliable_mark_active_outgoing (struct reliable *rel, struct buffer *buf, int opc
 	  e->opcode = opcode;
 	  e->next_try = 0;
 	  e->timeout = rel->initial_timeout;
-	  msg (D_REL_DEBUG, "ACK mark active outgoing ID " packet_id_format, (packet_id_print_type)e->packet_id);
+	  dmsg (D_REL_DEBUG, "ACK mark active outgoing ID " packet_id_format, (packet_id_print_type)e->packet_id);
 	  return;
 	}
     }
@@ -639,12 +664,11 @@ void
 reliable_ack_debug_print (const struct reliable_ack *ack, char *desc)
 {
   int i;
-  time_t current = time (NULL);
 
   printf ("********* struct reliable_ack %s\n", desc);
   for (i = 0; i < ack->len; ++i)
     {
-      printf ("  %d: " packet_id_format "\n", i, ack->packet_id[i]);
+      printf ("  %d: " packet_id_format "\n", i, (packet_id_print_type) ack->packet_id[i]);
     }
 }
 
@@ -652,12 +676,12 @@ void
 reliable_debug_print (const struct reliable *rel, char *desc)
 {
   int i;
-  time_t current = time (NULL);
+  update_time ();
 
   printf ("********* struct reliable %s\n", desc);
   printf ("  initial_timeout=%d\n", (int)rel->initial_timeout);
   printf ("  packet_id=" packet_id_format "\n", rel->packet_id);
-  printf ("  current=" time_format "\n", current);
+  printf ("  now=" time_format "\n", now);
   for (i = 0; i < rel->size; ++i)
     {
       const struct reliable_entry *e = &rel->array[i];
