@@ -5,7 +5,7 @@
  *             packet encryption, packet authentication, and
  *             packet compression.
  *
- *  Copyright (C) 2002-2003 James Yonan <jim@yonan.net>
+ *  Copyright (C) 2002-2004 James Yonan <jim@yonan.net>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -39,6 +39,8 @@
 #include "session_id.h"
 #include "mtu.h"
 
+/* #define EXPONENTIAL_BACKOFF */
+
 #define RELIABLE_ACK_SIZE 8
 
 struct reliable_ack
@@ -70,7 +72,7 @@ bool reliable_ack_write (struct reliable_ack *ack,
 			 const struct session_id *sid, int max, bool prepend);
 
 /* print a reliable ACK record coming off the wire */
-const char *reliable_ack_print(struct buffer* buf, bool verbose);
+const char *reliable_ack_print (struct buffer *buf, bool verbose, struct gc_arena *gc);
 
 /* add to extra_frame the maximum number of bytes we will need for reliable_ack_write */
 void reliable_ack_adjust_frame_parameters (struct frame* frame, int max);
@@ -115,7 +117,7 @@ void reliable_free (struct reliable *rel);
 bool reliable_empty (const struct reliable *rel);
 
 /* in how many seconds should we wake up to check for timeout */
-interval_t reliable_send_timeout (const struct reliable *rel, time_t current);
+interval_t reliable_send_timeout (const struct reliable *rel);
 
 /* del acknowledged items from send buf */
 void reliable_send_purge (struct reliable *rel, struct reliable_ack *ack);
@@ -139,13 +141,13 @@ struct buffer *reliable_get_buf_output_sequenced (struct reliable *rel);
 struct buffer *reliable_get_buf_sequenced (struct reliable *rel);
 
 /* return true if reliable_send would return a non-NULL result */
-bool reliable_can_send (const struct reliable *rel, time_t current);
+bool reliable_can_send (const struct reliable *rel);
 
 /* return next buffer to send to remote */
-struct buffer *reliable_send (struct reliable *rel, int *opcode, time_t current);
+struct buffer *reliable_send (struct reliable *rel, int *opcode);
 
 /* schedule all pending packets for immediate retransmit */
-void reliable_schedule_now (struct reliable *rel, time_t current);
+void reliable_schedule_now (struct reliable *rel);
 
 /* enable an incoming buffer previously returned by a get function as active */
 void reliable_mark_active_incoming (struct reliable *rel, struct buffer *buf,
