@@ -29,10 +29,10 @@
 #include "config.h"
 #endif
 
-#ifdef USE_PTHREAD
-
 #include "syshead.h"
 
+#ifdef USE_PTHREAD
+
 #include "thread.h"
 #include "buffer.h"
 #include "common.h"
@@ -41,28 +41,20 @@
 
 #include "memdbg.h"
 
-#if defined(USE_CRYPTO) && defined(USE_SSL)
-
-static pthread_mutex_t *ssl_lock_cs;
-static long *ssl_lock_count;
+static struct sparse_mutex *ssl_mutex;  /* GLOBAL */
 
 static void
 ssl_pthreads_locking_callback (int mode, int type, char *file, int line)
 {
-  msg (D_OPENSSL_LOCK, "SSL LOCK thread=%4lu mode=%s lock=%s %s:%d",
+  dmsg (D_OPENSSL_LOCK, "SSL LOCK thread=%4lu mode=%s lock=%s %s:%d",
 	   CRYPTO_thread_id (),
 	   (mode & CRYPTO_LOCK) ? "l" : "u",
 	   (type & CRYPTO_READ) ? "r" : "w", file, line);
 
   if (mode & CRYPTO_LOCK)
-    {
-      pthread_mutex_lock (&(ssl_lock_cs[type]));
-      ssl_lock_count[type]++;
-    }
+    pthread_mutex_lock (&ssl_mutex[type].mutex);
   else
-    {
-      pthread_mutex_unlock (&(ssl_lock_cs[type]));
-    }
+    pthread_mutex_unlock (&ssl_mutex[type].mutex);
 }
 
 static unsigned long
@@ -79,13 +71,11 @@ ssl_thread_setup (void)
 {
   int i;
 
-  ssl_lock_cs = OPENSSL_malloc (CRYPTO_num_locks () * sizeof (pthread_mutex_t));
-  ssl_lock_count = OPENSSL_malloc (CRYPTO_num_locks () * sizeof (long));
+#error L_MSG needs to be initialized as a recursive mutex
+
+  ssl_mutex = OPENSSL_malloc (CRYPTO_num_locks () * sizeof (struct sparse_mutex));
   for (i = 0; i < CRYPTO_num_locks (); i++)
-    {
-      ssl_lock_count[i] = 0;
-      pthread_mutex_init (&(ssl_lock_cs[i]), NULL);
-    }
+    pthread_mutex_init (&ssl_mutex[i].mutex, NULL);
 
   CRYPTO_set_id_callback ((unsigned long (*)(void)) ssl_pthreads_thread_id);
   CRYPTO_set_locking_callback ((void (*)(int, int, const char*, int)) ssl_pthreads_locking_callback);
@@ -96,86 +86,75 @@ ssl_thread_cleanup (void)
 {
   int i;
 
-  msg (D_OPENSSL_LOCK, "SSL LOCK cleanup");
+  dmsg (D_OPENSSL_LOCK, "SSL LOCK cleanup");
   CRYPTO_set_locking_callback (NULL);
   for (i = 0; i < CRYPTO_num_locks (); i++)
-    pthread_mutex_destroy (&(ssl_lock_cs[i]));
-  OPENSSL_free (ssl_lock_cs);
-  OPENSSL_free (ssl_lock_count);
+    pthread_mutex_destroy (&ssl_mutex[i].mutex);
+  OPENSSL_free (ssl_mutex);
 }
 
-#endif /* defined(USE_CRYPTO) && defined(USE_SSL) */
+struct sparse_mutex mutex_array[N_MUTEXES]; /* GLOBAL */
+bool pthread_initialized;                   /* GLOBAL */
 
-pthread_t x_main_thread_id;
-pthread_t x_work_thread_id;
-pthread_mutex_t x_lock_cs[N_MUTEXES];
-bool x_lock_cs_init;
-
-void
-work_thread_create (void *(*start_routine) (void *), void* arg)
+openvpn_thread_t
+openvpn_thread_create (void *(*start_routine) (void *), void* arg)
 {
-  ASSERT (x_main_thread_id);
-  ASSERT (!x_work_thread_id);
-  ASSERT (!pthread_create (&x_work_thread_id, NULL, start_routine, arg));
-  msg (D_THREAD_DEBUG, "CREATE THREAD ID=%lu", (unsigned long)x_work_thread_id);
+  openvpn_thread_t ret;
+  ASSERT (pthread_initialized);
+  ASSERT (!pthread_create (&ret, NULL, start_routine, arg));
+  dmsg (D_THREAD_DEBUG, "CREATE THREAD ID=%lu", (unsigned long)ret);
+  return ret;
 }
 
 void
-work_thread_join ()
+openvpn_thread_join (openvpn_thread_t id)
 {
-  if (x_work_thread_id)
-    {
-      pthread_join (x_work_thread_id, NULL);
-      x_work_thread_id = 0;
-    }
+  ASSERT (pthread_initialized);
+  pthread_join (id, NULL);
 }
 
 void
-thread_init ()
+openvpn_thread_init ()
 {
   int i;
 
-  ASSERT (!x_main_thread_id);
-  ASSERT (!x_work_thread_id);
+  ASSERT (!pthread_initialized);
 
   msg (M_INFO, "PTHREAD support initialized");
 
-  x_main_thread_id = pthread_self ();
-
   /* initialize OpenSSL library locking */
 #if defined(USE_CRYPTO) && defined(USE_SSL)
   ssl_thread_setup();
 #endif
   
   /* initialize static mutexes */
-  ASSERT (!x_lock_cs_init);
   for (i = 0; i < N_MUTEXES; i++)
-    ASSERT (!pthread_mutex_init (&(x_lock_cs[i]), NULL));
-  x_lock_cs_init = true;
+    ASSERT (!pthread_mutex_init (&mutex_array[i].mutex, NULL));
+
+  msg_thread_init ();
+
+  pthread_initialized = true;
 }
 
 void
-thread_cleanup ()
+openvpn_thread_cleanup ()
 {
-  ASSERT (!x_work_thread_id);
-  if (x_main_thread_id)
+  if (pthread_initialized)
     {
       int i;
 
+      pthread_initialized = false;
+
       /* cleanup OpenSSL library locking */
 #if defined(USE_CRYPTO) && defined(USE_SSL)
       ssl_thread_cleanup();
 #endif
 
       /* destroy static mutexes */
-      if (x_lock_cs_init)
-	{
-	  x_lock_cs_init = false;
-	  for (i = 0; i < N_MUTEXES; i++)
-	    ASSERT (!pthread_mutex_destroy (&(x_lock_cs[i])));
-	}
+      for (i = 0; i < N_MUTEXES; i++)
+	ASSERT (!pthread_mutex_destroy (&mutex_array[i].mutex));
 
-      x_main_thread_id = 0;
+      msg_thread_uninit ();
     }
 }
 
